@@ -51,6 +51,7 @@
 #define	PRI_DEBUG_Q931_ANOMALY 	(1 << 7)	/* Show unexpected events */
 #define PRI_DEBUG_APDU			(1 << 8)	/* Debug of APDU components such as ROSE */
 #define PRI_DEBUG_AOC			(1 << 9)	/* Debug of Advice of Charge ROSE Messages */
+#define PRI_DEBUG_CC			(1 << 10)	/* Debug call-completion. */
 
 #define PRI_DEBUG_ALL			(0xffff)	/* Everything */
 
@@ -97,6 +98,7 @@
 #define PRI_EVENT_RETRIEVE		24	/* RETRIEVE request received */
 #define PRI_EVENT_RETRIEVE_ACK	25	/* RETRIEVE_ACKNOWLEDGE received */
 #define PRI_EVENT_RETRIEVE_REJ	26	/* RETRIEVE_REJECT received */
+#define PRI_EVENT_CONNECT_ACK	27	/* CONNECT_ACKNOWLEDGE received */
 
 /* Simple states */
 #define PRI_STATE_DOWN		0
@@ -439,6 +441,14 @@ struct pri_party_subaddress {
 	unsigned char data[32];
 };
 
+/*! \brief Addressing information needed to identify an endpoint in a call. */
+struct pri_party_address {
+	/*! \brief Subscriber phone number */
+	struct pri_party_number number;
+	/*! \brief Subscriber subaddress */
+	struct pri_party_subaddress subaddress;
+};
+
 /*! \brief Information needed to identify an endpoint in a call. */
 struct pri_party_id {
 	/*! \brief Subscriber name */
@@ -503,11 +513,489 @@ struct pri_rerouting_data {
 	int invoke_id;
 };
 
-/* Subcommands derived from supplementary services. */
-#define PRI_SUBCMD_REDIRECTING		1
-#define PRI_SUBCMD_CONNECTED_LINE	2
-#define PRI_SUBCMD_REROUTING		3
+/*
+ * NOTE:
+ * The code surrounded by STATUS_REQUEST_PLACE_HOLDER is not implemented.
+ * The STATUS_REQUEST_PLACE_HOLDER code will be made unconditional if support
+ * for the messages is ever needed (and gets written).
+ */
 
+/* Subcommands derived from supplementary services. */
+#define PRI_SUBCMD_REDIRECTING				1	/*!< Redirecting information update */
+#define PRI_SUBCMD_CONNECTED_LINE			2	/*!< Connected line information update */
+#define PRI_SUBCMD_REROUTING				3	/*!< CallRerouting/CallDeflection received. */
+#if defined(STATUS_REQUEST_PLACE_HOLDER)
+#define PRI_SUBCMD_STATUS_REQ				4	/*!< Determine the status of the given party. */
+#define PRI_SUBCMD_STATUS_REQ_RSP			5	/*!< Status request response */
+#endif	/* defined(STATUS_REQUEST_PLACE_HOLDER) */
+#define PRI_SUBCMD_CC_AVAILABLE				6	/*!< Indicate that CC is available */
+#define PRI_SUBCMD_CC_REQ					7	/*!< CC activation request */
+#define PRI_SUBCMD_CC_REQ_RSP				8	/*!< CC activation request response */
+#define PRI_SUBCMD_CC_REMOTE_USER_FREE		9	/*!< Indicate that CC party B is available, party A is considered free. */
+#define PRI_SUBCMD_CC_B_FREE				10	/*!< Indicate that CC party B is available, party A is considered busy. */
+#define PRI_SUBCMD_CC_STATUS_REQ			11	/*!< Request/prod to receive updates of CC party A status */
+#define PRI_SUBCMD_CC_STATUS_REQ_RSP		12	/*!< Requested update of CC party A status */
+#define PRI_SUBCMD_CC_STATUS				13	/*!< Unsolicited update of CC party A status */
+#define PRI_SUBCMD_CC_CALL					14	/*!< Indicate that this call is a CC callback */
+#define PRI_SUBCMD_CC_CANCEL				15	/*!< Unsolicited indication that CC is canceled */
+#define PRI_SUBCMD_CC_STOP_ALERTING			16	/*!< Indicate that someone else has responed to remote user free */
+#define PRI_SUBCMD_TRANSFER_CALL			17	/*!< Request to transfer the specified calls together. */
+#define PRI_SUBCMD_AOC_S					18	/*!< Advice Of Charge Start information (Rate list) */
+#define PRI_SUBCMD_AOC_D					19	/*!< Advice Of Charge During information */
+#define PRI_SUBCMD_AOC_E					20	/*!< Advice Of Charge End information */
+#define PRI_SUBCMD_AOC_CHARGING_REQ			21	/*!< Advice Of Charge Request information */
+#define PRI_SUBCMD_AOC_CHARGING_REQ_RSP		22	/*!< Advice Of Charge Request Response information */
+#define PRI_SUBCMD_MCID_REQ					23	/*!< Malicious Call ID Request */
+#define PRI_SUBCMD_MCID_RSP					24	/*!< Malicious Call ID Request response */
+#define PRI_SUBCMD_DISPLAY_TEXT				25	/*!< Received display ie text */
+
+#if defined(STATUS_REQUEST_PLACE_HOLDER)
+struct pri_subcmd_status_request {
+	/*!
+	 * \brief Invoke id in case there are multiple outstanding requests.
+	 * \note Used to match any responses with the original invoke in case
+	 * there are several requests active.
+	 */
+	int invoke_id;
+	/*! \brief Party address requesting status about. */
+	struct pri_party_address party;
+};
+#endif	/* defined(STATUS_REQUEST_PLACE_HOLDER) */
+
+#if defined(STATUS_REQUEST_PLACE_HOLDER)
+struct pri_subcmd_status_request_rsp {
+	/*!
+	 * \brief Request id in case there are multiple outstanding requests.
+	 * \note Used to match any responses with the request in case there
+	 * are several requests active.
+	 */
+	int request_id;
+	/*!
+	 * \brief Response status to the status request.
+	 * \details
+	 * free(0),
+	 * busy(1),
+	 * incompatible(2)
+	 * timeout(3),
+	 */
+	int status;
+};
+#endif	/* defined(STATUS_REQUEST_PLACE_HOLDER) */
+
+struct pri_subcmd_cc_id {
+	/*! \brief Call-Completion record id */
+	long cc_id;
+};
+
+struct pri_subcmd_cc_request {
+	/*! \brief Call-Completion record id */
+	long cc_id;
+	/*!
+	 * \brief Mode of call-completion requested.
+	 * \details
+	 * ccbs(0),
+	 * ccnr(1)
+	 */
+	int mode;
+};
+
+struct pri_subcmd_cc_request_rsp {
+	/*! \brief Call-Completion record id */
+	long cc_id;
+	/*!
+	 * \brief Status of the requested call-completion activation.
+	 * \details
+	 * success(0),
+	 * timeout(1),
+	 * error(2),
+	 * reject(3)
+	 */
+	int status;
+	/*!
+	 * \brief Failure code that can be converted to a string to further
+	 * explain the non-timeout failure.
+	 * \note Valid when status is error or reject.
+	 * \note Use pri_facility_error2str() to convert the error_code.
+	 * \note Use pri_facility_reject2str() to convert the reject_code.
+	 */
+	int fail_code;
+};
+
+struct pri_subcmd_cc_status {
+	/*! \brief Call-Completion record id */
+	long cc_id;
+	/*!
+	 * \brief Party A status.
+	 * \details
+	 * free(0),
+	 * busy(1)
+	 */
+	int status;
+};
+
+struct pri_subcmd_cc_cancel {
+	/*! \brief Call-Completion record id */
+	long cc_id;
+	/*!
+	 * \brief TRUE if the cc_id is for an agent.
+	 * \note This is a convenience value so the upper layer can know which
+	 * list it should search for the cc_id.
+	 */
+	int is_agent;
+};
+
+struct pri_subcmd_transfer {
+	/*! \brief Opaque call pointer for transfer with other call. */
+	q931_call *call_1;
+	/*! \brief Opaque call pointer for transfer with other call. */
+	q931_call *call_2;
+	/*! TRUE if call_1 is on hold. */
+	int is_call_1_held;
+	/*! TRUE if call_2 is on hold. */
+	int is_call_2_held;
+	/*! Invocation ID to use when sending a reply to the transfer request. */
+	int invoke_id;
+};
+
+/*! \brief What is being charged. */
+enum PRI_AOC_CHARGED_ITEM {
+	PRI_AOC_CHARGED_ITEM_NOT_AVAILABLE,
+	PRI_AOC_CHARGED_ITEM_SPECIAL_ARRANGEMENT,
+	PRI_AOC_CHARGED_ITEM_BASIC_COMMUNICATION,
+	PRI_AOC_CHARGED_ITEM_CALL_ATTEMPT,
+	PRI_AOC_CHARGED_ITEM_CALL_SETUP,
+	PRI_AOC_CHARGED_ITEM_USER_USER_INFO,
+	PRI_AOC_CHARGED_ITEM_SUPPLEMENTARY_SERVICE,
+};
+
+/*! \brief Rate method being used. */
+enum PRI_AOC_RATE_TYPE {
+	PRI_AOC_RATE_TYPE_NOT_AVAILABLE,
+	PRI_AOC_RATE_TYPE_FREE,
+	PRI_AOC_RATE_TYPE_FREE_FROM_BEGINNING,
+	PRI_AOC_RATE_TYPE_DURATION,
+	PRI_AOC_RATE_TYPE_FLAT,
+	PRI_AOC_RATE_TYPE_VOLUME,
+	PRI_AOC_RATE_TYPE_SPECIAL_CODE,
+};
+
+enum PRI_AOC_TIME_SCALE {
+	PRI_AOC_TIME_SCALE_HUNDREDTH_SECOND,
+	PRI_AOC_TIME_SCALE_TENTH_SECOND,
+	PRI_AOC_TIME_SCALE_SECOND,
+	PRI_AOC_TIME_SCALE_TEN_SECOND,
+	PRI_AOC_TIME_SCALE_MINUTE,
+	PRI_AOC_TIME_SCALE_HOUR,
+	PRI_AOC_TIME_SCALE_DAY,
+};
+
+struct pri_aoc_time {
+	/*! LengthOfTimeUnit (Not valid if length is zero.) */
+	long length;
+	/*! \see enum PRI_AOC_TIME_SCALE */
+	int scale;
+};
+
+enum PRI_AOC_MULTIPLIER {
+	PRI_AOC_MULTIPLIER_THOUSANDTH,
+	PRI_AOC_MULTIPLIER_HUNDREDTH,
+	PRI_AOC_MULTIPLIER_TENTH,
+	PRI_AOC_MULTIPLIER_ONE,
+	PRI_AOC_MULTIPLIER_TEN,
+	PRI_AOC_MULTIPLIER_HUNDRED,
+	PRI_AOC_MULTIPLIER_THOUSAND,
+};
+
+struct pri_aoc_amount {
+	long cost;
+	/*! \see enum PRI_AOC_MULTIPLIER */
+	int multiplier;
+};
+
+struct pri_aoc_duration {
+	struct pri_aoc_amount amount;
+	struct pri_aoc_time time;
+	/*! Not present if the granularity time is zero. */
+	struct pri_aoc_time granularity;
+	/*!
+	 * \brief Charging interval type
+	 * \details
+	 * continuousCharging(0),
+	 * stepFunction(1)
+	 */
+	int charging_type;
+	/*! Name of currency involved.  Null terminated. */
+	char currency[10 + 1];
+};
+
+struct pri_aoc_flat {
+	struct pri_aoc_amount amount;
+	/*! Name of currency involved.  Null terminated. */
+	char currency[10 + 1];
+};
+
+enum PRI_AOC_VOLUME_UNIT {
+	PRI_AOC_VOLUME_UNIT_OCTET,
+	PRI_AOC_VOLUME_UNIT_SEGMENT,
+	PRI_AOC_VOLUME_UNIT_MESSAGE,
+};
+
+struct pri_aoc_volume {
+	struct pri_aoc_amount amount;
+	/*! \see enum PRI_AOC_VOLUME_UNIT */
+	int unit;
+	/*! Name of currency involved.  Null terminated. */
+	char currency[10 + 1];
+};
+
+struct pri_aoc_s_element {
+	/*!
+	 * \brief What is being charged.
+	 * \see enum PRI_AOC_CHARGED_ITEM
+	 */
+	int chargeable;
+	/*!
+	 * \brief Rate method being used.
+	 * \see enum PRI_AOC_RATE_TYPE
+	 */
+	int rate_type;
+	/*! \brief Charge rate being applied. */
+	union {
+		struct pri_aoc_duration duration;
+		struct pri_aoc_flat flat;
+		struct pri_aoc_volume volume;
+		int special;
+	} rate;
+};
+
+struct pri_subcmd_aoc_s {
+	/*!
+	 * \brief Number of items in the rate list.
+	 * \note If the list is empty then the charging information is not available.
+	 */
+	int num_items;
+	struct pri_aoc_s_element item[10];
+};
+
+enum PRI_AOC_DE_CHARGE {
+	PRI_AOC_DE_CHARGE_NOT_AVAILABLE,
+	PRI_AOC_DE_CHARGE_FREE,
+	PRI_AOC_DE_CHARGE_CURRENCY,
+	PRI_AOC_DE_CHARGE_UNITS,
+};
+
+struct pri_aoc_recorded_currency {
+	struct pri_aoc_amount amount;
+	/*! Name of currency involved.  Null terminated. */
+	char currency[10 + 1];
+};
+
+struct pri_aoc_units_element {
+	/*! Number of units recorded. -1 if not available. */
+	long number;
+	/*! Type of unit recorded. -1 if not available. */
+	int type;
+};
+
+struct pri_aoc_recorded_units {
+	int num_items;
+	struct pri_aoc_units_element item[32];
+};
+
+enum PRI_AOC_D_BILLING_ID {
+	PRI_AOC_D_BILLING_ID_NOT_AVAILABLE,
+	PRI_AOC_D_BILLING_ID_NORMAL,
+	PRI_AOC_D_BILLING_ID_REVERSE,
+	PRI_AOC_D_BILLING_ID_CREDIT_CARD,
+};
+struct pri_subcmd_aoc_d {
+	/*!
+	 * \brief What is being charged.
+	 * \see enum PRI_AOC_DE_CHARGE
+	 */
+	int charge;
+	/*!
+	 * \brief Billing accumulation
+	 * \details
+	 * subTotal(0),
+	 * total(1)
+	 */
+	int billing_accumulation;
+	/*! \see enum PRI_AOC_D_BILLING_ID */
+	int billing_id;
+	union {
+		/*! Recorded currency */
+		struct pri_aoc_recorded_currency money;
+		/*! Recorded units list */
+		struct pri_aoc_recorded_units unit;
+	} recorded;
+};
+
+enum PRI_AOC_E_BILLING_ID {
+	PRI_AOC_E_BILLING_ID_NOT_AVAILABLE,
+	PRI_AOC_E_BILLING_ID_NORMAL,
+	PRI_AOC_E_BILLING_ID_REVERSE,
+	PRI_AOC_E_BILLING_ID_CREDIT_CARD,
+	PRI_AOC_E_BILLING_ID_CALL_FORWARDING_UNCONDITIONAL,
+	PRI_AOC_E_BILLING_ID_CALL_FORWARDING_BUSY,
+	PRI_AOC_E_BILLING_ID_CALL_FORWARDING_NO_REPLY,
+	PRI_AOC_E_BILLING_ID_CALL_DEFLECTION,
+	PRI_AOC_E_BILLING_ID_CALL_TRANSFER,
+};
+
+enum PRI_AOC_E_CHARGING_ASSOCIATION {
+	PRI_AOC_E_CHARGING_ASSOCIATION_NOT_AVAILABLE,
+	PRI_AOC_E_CHARGING_ASSOCIATION_NUMBER,
+	PRI_AOC_E_CHARGING_ASSOCIATION_ID,
+};
+
+struct pri_aoc_e_charging_association {
+	union {
+		/*! Charged number */
+		struct pri_party_number number;
+		/*! Charge identifier */
+		int id;
+	} charge;
+	/*! \see enum PRI_AOC_E_CHARGING_ASSOCIATION */
+	int charging_type;
+};
+
+struct pri_subcmd_aoc_e {
+	/*!
+	 * \brief What is being charged.
+	 * \see enum PRI_AOC_DE_CHARGE
+	 */
+	int charge;
+	/*! \see enum PRI_AOC_E_BILLING_ID */
+	int billing_id;
+	union {
+		/*! Recorded currency */
+		struct pri_aoc_recorded_currency money;
+		/*! Recorded units list */
+		struct pri_aoc_recorded_units unit;
+	} recorded;
+	/*! Charging association. */
+	struct pri_aoc_e_charging_association associated;
+};
+
+enum PRI_AOC_REQ_RSP {
+	/* Error Results */
+	PRI_AOC_REQ_RSP_ERROR_NOT_IMPLEMENTED,
+	PRI_AOC_REQ_RSP_ERROR_NOT_AVAILABLE,
+	PRI_AOC_REQ_RSP_ERROR_TIMEOUT,
+	PRI_AOC_REQ_RSP_ERROR_REJECT,
+	/* generic error result all other errors are lumped into */
+	PRI_AOC_REQ_RSP_ERROR,
+
+	/* AOC Results */
+	PRI_AOC_REQ_RSP_CHARGING_INFO_FOLLOWS,
+	PRI_AOC_REQ_RSP_CURRENCY_INFO_LIST,
+	PRI_AOC_REQ_RSP_SPECIAL_ARR,
+};
+
+enum PRI_AOC_REQUEST {
+	PRI_AOC_REQUEST_S = (1 << 0),
+	PRI_AOC_REQUEST_D = (1 << 1),
+	PRI_AOC_REQUEST_E = (1 << 2),
+};
+
+struct pri_subcmd_aoc_request_response {
+	/*!
+	 * \brief aoc_s data from response
+	 */
+	struct pri_subcmd_aoc_s aoc_s;
+
+	/*!
+	 * \brief if the aoc_s msg is present, this will be set
+	 */
+	int valid_aoc_s;
+
+	/*!
+	 * \brief What type of aoc was requested.
+	 * \see enum PRI_AOC_REQUEST
+	 */
+	int charging_request;
+
+	/*!
+	 * \brief response to the charging_request
+	 * \see enum PRI_AOC_REQ_RSP
+	 */
+	int charging_response;
+};
+
+struct pri_subcmd_aoc_request {
+	/*!
+	 * \brief What types of aoc are being requested.
+	 * \see enum PRI_AOC_REQUEST
+	 */
+	int charging_request;
+
+	/*!
+	 * \brief Value given by the initiating request.
+	 */
+	 int invoke_id;
+};
+
+struct pri_subcmd_mcid_req {
+	/*!
+	 * \brief Information libpri knows about the malicious caller.
+	 * \note For the convenience of the upper layer.  This information
+	 * may be incomplete if the upper layer redacted some caller
+	 * information because it was restricted.
+	 */
+	struct pri_party_id originator;
+	/*! \brief Information libpri knows about the callee. */
+	struct pri_party_id answerer;
+};
+
+struct pri_subcmd_mcid_rsp {
+	/*!
+	 * \brief MCID request response status.
+	 * \details
+	 * success(0),
+	 * timeout(1),
+	 * error(2),
+	 * reject(3)
+	 */
+	int status;
+	/*!
+	 * \brief Failure code that can be converted to a string to further
+	 * explain the non-timeout failure.
+	 * \note Valid when status is error or reject.
+	 * \note Use pri_facility_error2str() to convert the error_code.
+	 * \note Use pri_facility_reject2str() to convert the reject_code.
+	 */
+	int fail_code;
+};
+
+struct pri_subcmd_display_txt {
+	/*!
+	 * \brief Character set the text is using.
+	 * \details
+	 * unknown(0),
+	 * iso8859-1(1),
+	 * enum-value-withdrawn-by-ITU-T(2)
+	 * iso8859-2(3),
+	 * iso8859-3(4),
+	 * iso8859-4(5),
+	 * iso8859-5(6),
+	 * iso8859-7(7),
+	 * iso10646-BmpString(8),
+	 * iso10646-utf-8String(9)
+	 */
+	int char_set;
+	/*!
+	 * \brief Number of octets in the display message.
+	 * \note Not including any added null terminator.
+	 */
+	int length;
+	/*!
+	 * \brief Display text data.
+	 * \note Null terminated on receive.
+	 * \note Does not need to be null terminated on send.
+	 */
+	char text[128];
+};
 
 struct pri_subcommand {
 	/*! PRI_SUBCMD_xxx defined values */
@@ -518,6 +1006,30 @@ struct pri_subcommand {
 		struct pri_party_connected_line connected_line;
 		struct pri_party_redirecting redirecting;
 		struct pri_rerouting_data rerouting;
+#if defined(STATUS_REQUEST_PLACE_HOLDER)
+		struct pri_subcmd_status_request status_request;
+		struct pri_subcmd_status_request_rsp status_request_rsp;
+#endif	/* defined(STATUS_REQUEST_PLACE_HOLDER) */
+		struct pri_subcmd_cc_id cc_available;
+		struct pri_subcmd_cc_request cc_request;
+		struct pri_subcmd_cc_request_rsp cc_request_rsp;
+		struct pri_subcmd_cc_id cc_remote_user_free;
+		struct pri_subcmd_cc_id cc_b_free;
+		struct pri_subcmd_cc_id cc_stop_alerting;
+		struct pri_subcmd_cc_id cc_status_req;
+		struct pri_subcmd_cc_status cc_status_req_rsp;
+		struct pri_subcmd_cc_status cc_status;
+		struct pri_subcmd_cc_id cc_call;
+		struct pri_subcmd_cc_cancel cc_cancel;
+		struct pri_subcmd_transfer transfer;
+		struct pri_subcmd_aoc_request aoc_request;
+		struct pri_subcmd_aoc_request_response aoc_request_response;
+		struct pri_subcmd_aoc_s aoc_s;
+		struct pri_subcmd_aoc_d aoc_d;
+		struct pri_subcmd_aoc_e aoc_e;
+		struct pri_subcmd_mcid_req mcid_req;
+		struct pri_subcmd_mcid_rsp mcid_rsp;
+		struct pri_subcmd_display_txt display;
 	} u;
 };
 
@@ -662,7 +1174,7 @@ typedef struct pri_event_hangup {
 	int cause;
 	int cref;
 	q931_call *call;			/* Opaque call pointer of call hanging up. */
-	long aoc_units;				/* Advise of Charge number of charged units */
+	long aoc_units;				/* Advice of Charge number of charged units */
 	char useruserinfo[260];		/* User->User info */
 	struct pri_subcommands *subcmds;
 	/*!
@@ -776,6 +1288,13 @@ struct pri_event_retrieve_rej {
 	struct pri_subcommands *subcmds;
 };
 
+struct pri_event_connect_ack {
+	int e;
+	int channel;
+	q931_call *call;
+	struct pri_subcommands *subcmds;
+};
+
 typedef union {
 	int e;
 	pri_event_generic gen;		/* Generic view */
@@ -800,6 +1319,7 @@ typedef union {
 	struct pri_event_retrieve retrieve;
 	struct pri_event_retrieve_ack retrieve_ack;
 	struct pri_event_retrieve_rej retrieve_rej;
+	struct pri_event_connect_ack connect_ack;
 } pri_event;
 
 struct pri;
@@ -819,6 +1339,9 @@ struct pri *pri_new_bri(int fd, int ptpmode, int nodetype, int switchtype);
 /* Create D-channel just as above with user defined I/O callbacks and data */
 struct pri *pri_new_cb(int fd, int nodetype, int switchtype, pri_io_cb io_read, pri_io_cb io_write, void *userdata);
 
+/* Create BRI D-channel just as above with user defined I/O callbacks and data */
+struct pri *pri_new_bri_cb(int fd, int ptpmode, int nodetype, int switchtype, pri_io_cb io_read, pri_io_cb io_write, void *userdata);
+
 /* Retrieve the user data associated with the D channel */
 void *pri_get_userdata(struct pri *pri);
 
@@ -867,6 +1390,24 @@ char *pri_plan2str(int plan);
 /* Turn cause into a string */
 char *pri_cause2str(int cause);
 
+/*!
+ * \brief Convert the given facility error code to a descriptive string.
+ *
+ * \param facility_error_code Error code to convert to a string.
+ *
+ * \return Descriptive error string.
+ */
+const char *pri_facility_error2str(int facility_error_code);
+
+/*!
+ * \brief Convert the given facility reject code to a descriptive string.
+ *
+ * \param facility_reject_code Error code to convert to a string.
+ *
+ * \return Descriptive reject string.
+ */
+const char *pri_facility_reject2str(int facility_reject_code);
+
 /* Acknowledge a call and place it on the given channel.  Set info to non-zero if there
    is in-band data available on the channel */
 int pri_acknowledge(struct pri *pri, q931_call *call, int channel, int info);
@@ -882,10 +1423,33 @@ int pri_keypad_facility(struct pri *pri, q931_call *call, const char *digits);
    Set non-isdn to non-zero if you are not connecting to ISDN equipment */
 int pri_need_more_info(struct pri *pri, q931_call *call, int channel, int nonisdn);
 
-/* Answer the call on the given channel (ignored if you called acknowledge already).
+/* Answer(CONNECT) the call on the given channel.
    Set non-isdn to non-zero if you are not connecting to ISDN equipment */
 int pri_answer(struct pri *pri, q931_call *call, int channel, int nonisdn);
 
+/*!
+ * \brief Send the manual CONNECT_ACKNOWLEDGE message.
+ *
+ * \param ctrl D channel controller.
+ * \param call Q.931 call leg.
+ * \param channel Selected channel to assign to the call waiting call.
+ * Zero if do not include the channel id ie in the CONNECT_ACKNOWLEDGE message.
+ *
+ * \retval 0 on success.
+ * \retval -1 on error.
+ */
+int pri_connect_ack(struct pri *ctrl, q931_call *call, int channel);
+
+/*!
+ * \brief Set the manual CONNECT_ACKNOWLEDGE message enable flag.
+ *
+ * \param ctrl D channel controller.
+ * \param enable TRUE to enable manual CONNECT_ACKNOWLEDGE message feature.
+ *
+ * \return Nothing
+ */
+void pri_connect_ack_enable(struct pri *ctrl, int enable);
+
 /*!
  * \brief Give connected line information to a call
  * \note Could be used instead of pri_sr_set_caller_party() before calling pri_setup().
@@ -1053,12 +1617,52 @@ int pri_mwi_activate(struct pri *pri, q931_call *c, char *caller, int callerplan
 /* Send an MWI deactivate request to a remote location */
 int pri_mwi_deactivate(struct pri *pri, q931_call *c, char *caller, int callerplan, char *callername, int callerpres, char *called, int calledplan);
 
+/*!
+ * \brief Send a MWI indication on the specified D channel.
+ *
+ * \param ctrl D channel controller.
+ * \param mailbox Party receiving notification.
+ * \param basic_service Basic service enum (-1 if not present).
+ * \param num_messages NumberOfMessages (-1 if not present).
+ * \param caller_id Party leaving message (NULL if not present).
+ * \param timestamp When message left. (Generalized Time format, NULL if not present)
+ * \param message_reference Message reference number (-1 if not present).
+ * \param message_status Message status: added(0), removed(1).
+ *
+ * \retval 0 on success.
+ * \retval -1 on error.
+ */
+int pri_mwi_indicate(struct pri *ctrl, const struct pri_party_id *mailbox,
+	int basic_service, int num_messages, const struct pri_party_id *caller_id,
+	const char *timestamp, int message_reference, int message_status);
+
+/*!
+ * \brief Send a MWI indication on the specified D channel. (Take two)
+ *
+ * \param ctrl D channel controller.
+ * \param mailbox Party receiving notification.
+ * \param vm_id Voicemail system number (NULL if not present).
+ * \param basic_service Basic service enum (-1 if not present).
+ * \param num_messages NumberOfMessages (-1 if not present).
+ * \param caller_id Party leaving message (NULL if not present).
+ * \param timestamp When message left. (Generalized Time format, NULL if not present)
+ * \param message_reference Message reference number (-1 if not present).
+ * \param message_status Message status: added(0), removed(1).
+ *
+ * \retval 0 on success.
+ * \retval -1 on error.
+ */
+int pri_mwi_indicate_v2(struct pri *ctrl, const struct pri_party_id *mailbox,
+	const struct pri_party_id *vm_id, int basic_service, int num_messages,
+	const struct pri_party_id *caller_id, const char *timestamp, int message_reference,
+	int message_status);
+
 /* Set service message support flag */
 int pri_set_service_message_support(struct pri *pri, int supportflag);
 
 #define PRI_2BCT
 /* Attempt to pass the channels back to the NET side if compatable and
- * suscribed.  Sometimes called 2 bchannel transfer (2BCT) */
+ * subscribed.  Sometimes called 2 bchannel transfer (2BCT) */
 int pri_channel_bridge(q931_call *call1, q931_call *call2);
 
 /* Override message and error stuff */
@@ -1098,6 +1702,79 @@ void pri_set_inbanddisconnect(struct pri *pri, unsigned int enable);
    (and maybe some timers) */
 void pri_enslave(struct pri *master, struct pri *slave);
 
+/*!
+ * \brief  Request AOC on call setup
+ *
+ * \param call setup struct to set charging request info on
+ * \param charging request to set on setup struct
+ *
+ * \retval 0 on success
+ * \retval -1 on failure
+ */
+int pri_sr_set_aoc_charging_request(struct pri_sr *sr, int charging_request);
+
+/*!
+ * \brief Send AOC Request Response to a request for AOC-S
+ *
+ * \param ctrl D channel controller.
+ * \param call Q.931 call leg.
+ * \param invoke_id for response message
+ * \param aoc_s message for response
+ *
+ * \retval 0 on success
+ * \retval -1 on failure
+ */
+int pri_aoc_s_request_response_send(struct pri *ctrl, q931_call *call, int invoke_id, const struct pri_subcmd_aoc_s *aoc_s);
+
+/*!
+ * \brief Send AOC Request Response to a request for AOC-D or AOC-E
+ *
+ * \param ctrl D channel controller.
+ * \param call Q.931 call leg.
+ * \param response in form of enum PRI_AOC_REQ_RSP
+ * \param invoke_id for response message
+ *
+ * \retval 0 on success
+ * \retval -1 on failure
+ */
+int pri_aoc_de_request_response_send(struct pri *ctrl, q931_call *call, int response, int invoke_id);
+
+/*!
+ * \brief Send AOC-S message.
+ *
+ * \param ctrl D channel controller.
+ * \param call Q.931 call leg.
+ * \param aoc_s message to send
+ *
+ * \retval 0 on success
+ * \retval -1 on failure
+ */
+int pri_aoc_s_send(struct pri *ctrl, q931_call *c, const struct pri_subcmd_aoc_s *aoc_s);
+
+/*!
+ * \brief Send AOC-D message.
+ *
+ * \param ctrl D channel controller.
+ * \param call Q.931 call leg.
+ * \param aoc_d message to send
+ *
+ * \retval 0 on success
+ * \retval -1 on failure
+ */
+int pri_aoc_d_send(struct pri *ctrl, q931_call *c, const struct pri_subcmd_aoc_d *aoc_d);
+
+/*!
+ * \brief Send AOC-E message.
+ *
+ * \param ctrl D channel controller.
+ * \param call Q.931 call leg.
+ * \param aoc_e message to send
+ *
+ * \retval 0 on success
+ * \retval -1 on failure
+ */
+int pri_aoc_e_send(struct pri *ctrl, q931_call *c, const struct pri_subcmd_aoc_e *aoc_e);
+
 #define PRI_GR303_SUPPORT
 #define PRI_ENSLAVE_SUPPORT
 #define PRI_SETUP_CALL
@@ -1182,6 +1859,88 @@ enum PRI_REROUTING_RSP_CODE {
  */
 int pri_rerouting_rsp(struct pri *ctrl, q931_call *call, int invoke_id, enum PRI_REROUTING_RSP_CODE code);
 
+/*!
+ * \brief Set the call transfer feature enable flag.
+ *
+ * \param ctrl D channel controller.
+ * \param enable TRUE to enable call transfer feature.
+ *
+ * \return Nothing
+ */
+void pri_transfer_enable(struct pri *ctrl, int enable);
+
+/*!
+ * \brief Send the call transfer response message.
+ *
+ * \param ctrl D channel controller.
+ * \param call Q.931 call leg.
+ * \param invoke_id Value given by the initiating request.
+ * \param is_successful TRUE if the transfer was successful.
+ *
+ * \retval 0 on success.
+ * \retval -1 on error.
+ */
+int pri_transfer_rsp(struct pri *ctrl, q931_call *call, int invoke_id, int is_successful);
+
+/*!
+ * \brief Set the advice of charge events feature enable flag.
+ *
+ * \param ctrl D channel controller.
+ * \param enable TRUE to enable AOC events feature.
+ *
+ * \return Nothing
+ */
+void pri_aoc_events_enable(struct pri *ctrl, int enable);
+
+#define PRI_DISPLAY_OPTION_BLOCK		(1 << 0)	/*!< Do not pass display text. */
+#define PRI_DISPLAY_OPTION_NAME_INITIAL	(1 << 1)	/*!< Use display in SETUP/CONNECT for name. */
+#define PRI_DISPLAY_OPTION_NAME_UPDATE	(1 << 2)	/*!< Use display in FACILITY/NOTIFY for COLP name if appropriate. */
+#define PRI_DISPLAY_OPTION_TEXT			(1 << 3)	/*!< Pass arbitrary display text in INFORMATION messages during call. */
+
+/*!
+ * \brief Set the display ie send policy options.
+ *
+ * \param ctrl D channel controller.
+ * \param flags Option flags to apply.
+ *
+ * \note
+ * If no flags set then legacy default behaviour.
+ *
+ * \note
+ * Not all options are supported by all switches.
+ *
+ * \return Nothing
+ */
+void pri_display_options_send(struct pri *ctrl, unsigned long flags);
+
+/*!
+ * \brief Set the display ie receive policy options.
+ *
+ * \param ctrl D channel controller.
+ * \param flags Option flags to apply.
+ *
+ * \note
+ * If no flags set then legacy default behaviour.
+ *
+ * \note
+ * Not all options are supported by all switches.
+ *
+ * \return Nothing
+ */
+void pri_display_options_receive(struct pri *ctrl, unsigned long flags);
+
+/*!
+ * \brief Send display text during a call.
+ *
+ * \param ctrl D channel controller.
+ * \param call Q.931 call leg
+ * \param display Display text to send.
+ *
+ * \retval 0 on success.
+ * \retval -1 on error.
+ */
+int pri_display_text(struct pri *ctrl, q931_call *call, const struct pri_subcmd_display_txt *display);
+
 /*!
  * \brief Set the call hold feature enable flag.
  *
@@ -1262,6 +2021,108 @@ int pri_retrieve_ack(struct pri *ctrl, q931_call *call, int channel);
  */
 int pri_retrieve_rej(struct pri *ctrl, q931_call *call, int cause);
 
+#if defined(STATUS_REQUEST_PLACE_HOLDER)
+int pri_status_req(struct pri *ctrl, int request_id, const struct pri_sr *req);
+void pri_status_req_rsp(struct pri *ctrl, int invoke_id, int status);
+#endif	/* defined(STATUS_REQUEST_PLACE_HOLDER) */
+
+/*!
+ * \brief Set the Malicious Call ID feature enable flag.
+ *
+ * \param ctrl D channel controller.
+ * \param enable TRUE to enable MCID feature.
+ *
+ * \return Nothing
+ */
+void pri_mcid_enable(struct pri *ctrl, int enable);
+
+/*!
+ * \brief Send the MCID request message.
+ *
+ * \param ctrl D channel controller.
+ * \param call Q.931 call leg
+ *
+ * \retval 0 on success.  You should wait for a PRI_SUBCMD_MCID_RSP
+ * to continue clearing the call if it was in progress.
+ * \retval -1 on error.
+ */
+int pri_mcid_req_send(struct pri *ctrl, q931_call *call);
+
+/*!
+ * \brief Set the call completion feature enable flag.
+ *
+ * \param ctrl D channel controller.
+ * \param enable TRUE to enable call completion feature.
+ *
+ * \return Nothing
+ */
+void pri_cc_enable(struct pri *ctrl, int enable);
+
+/*!
+ * \brief Set the PTMP NT call completion recall mode.
+ *
+ * \param ctrl D channel controller.
+ * \param mode globalRecall(0), specificRecall(1)
+ *
+ * \return Nothing
+ */
+void pri_cc_recall_mode(struct pri *ctrl, int mode);
+
+/*!
+ * \brief Set the Q.SIG call completion signaling link retention mode.
+ * (Requestor/Initiator/Originator/Party-A)
+ *
+ * \param ctrl D channel controller.
+ * \param signaling_retention release(0), retain(1), do-not-care(2).
+ *
+ * \return Nothing
+ */
+void pri_cc_retain_signaling_req(struct pri *ctrl, int signaling_retention);
+
+/*!
+ * \brief Set the Q.SIG call completion signaling link retention mode.
+ * (Responder/Answerer/Party-B)
+ *
+ * \param ctrl D channel controller.
+ * \param signaling_retention release(0), retain(1).
+ *
+ * \return Nothing
+ */
+void pri_cc_retain_signaling_rsp(struct pri *ctrl, int signaling_retention);
+
+long pri_cc_available(struct pri *ctrl, q931_call *call);
+int pri_cc_req(struct pri *ctrl, long cc_id, int mode);
+int pri_cc_req_rsp(struct pri *ctrl, long cc_id, int status);
+void pri_cc_remote_user_free(struct pri *ctrl, long cc_id);
+void pri_cc_b_free(struct pri *ctrl, long cc_id);
+void pri_cc_stop_alerting(struct pri *ctrl, long cc_id);
+void pri_cc_status_req(struct pri *ctrl, long cc_id);
+void pri_cc_status_req_rsp(struct pri *ctrl, long cc_id, int status);
+void pri_cc_status(struct pri *ctrl, long cc_id, int status);
+int pri_cc_call(struct pri *ctrl, long cc_id, q931_call *call, struct pri_sr *req);
+void pri_cc_cancel(struct pri *ctrl, long cc_id);
+
+/* Date/time ie send policy option values. */
+#define PRI_DATE_TIME_SEND_DEFAULT		0	/*!< Send date/time ie default. */
+#define PRI_DATE_TIME_SEND_NO			1	/*!< Send date/time ie never. */
+#define PRI_DATE_TIME_SEND_DATE			2	/*!< Send date/time ie date only. */
+#define PRI_DATE_TIME_SEND_DATE_HH		3	/*!< Send date/time ie date and hour. */
+#define PRI_DATE_TIME_SEND_DATE_HHMM	4	/*!< Send date/time ie date, hour, and minute. */
+#define PRI_DATE_TIME_SEND_DATE_HHMMSS	5	/*!< Send date/time ie date, hour, minute, and second. */
+
+/*!
+ * \brief Set the date/time ie send policy option.
+ *
+ * \param ctrl D channel controller.
+ * \param option Policy option to set.
+ *
+ * \note
+ * Only valid in NT mode.
+ *
+ * \return Nothing
+ */
+void pri_date_time_send_option(struct pri *ctrl, int option);
+
 /* Get/Set PRI Timers  */
 #define PRI_GETSET_TIMERS
 int pri_set_timer(struct pri *pri, int timer, int value);
@@ -1309,6 +2170,33 @@ enum PRI_TIMERS_AND_COUNTERS {
 
 	PRI_TIMER_T_RESPONSE,	/*!< Maximum time to wait for a typical APDU response. */
 
+	PRI_TIMER_T_STATUS,		/*!< Max time to wait for all replies to check for compatible terminals */
+
+	PRI_TIMER_T_ACTIVATE,	/*!< Request supervision timeout. */
+	PRI_TIMER_T_DEACTIVATE,	/*!< Deactivate supervision timeout. */
+	PRI_TIMER_T_INTERROGATE,/*!< Interrogation supervision timeout. */
+
+	/* ETSI call-completion timers */
+	PRI_TIMER_T_RETENTION,	/*!< Max time to wait for user A to activate call-completion. */
+	PRI_TIMER_T_CCBS1,		/*!< T-STATUS timer equivalent for CC user A status. */
+	PRI_TIMER_T_CCBS2,		/*!< Max time the CCBS service will be active */
+	PRI_TIMER_T_CCBS3,		/*!< Max time to wait for user A to respond to user B availability. */
+	PRI_TIMER_T_CCBS4,		/*!< CC user B guard time before sending CC recall indication. */
+	PRI_TIMER_T_CCBS5,		/*!< Network B CCBS supervision timeout. */
+	PRI_TIMER_T_CCBS6,		/*!< Network A CCBS supervision timeout. */
+	PRI_TIMER_T_CCNR2,		/*!< Max time the CCNR service will be active */
+	PRI_TIMER_T_CCNR5,		/*!< Network B CCNR supervision timeout. */
+	PRI_TIMER_T_CCNR6,		/*!< Network A CCNR supervision timeout. */
+
+	/* Q.SIG call-completion timers */
+	PRI_TIMER_QSIG_CC_T1,	/*!< CC request supervision timeout. */
+	PRI_TIMER_QSIG_CCBS_T2,	/*!< CCBS supervision timeout. */
+	PRI_TIMER_QSIG_CCNR_T2,	/*!< CCNR supervision timeout. */
+	PRI_TIMER_QSIG_CC_T3,	/*!< Max time to wait for user A to respond to user B availability. */
+	PRI_TIMER_QSIG_CC_T4,	/*!< Path reservation supervision timeout. */
+
+	PRI_TIMER_T312,			/*!< Supervise broadcast SETUP message call reference retention. */
+
 	/* Must be last in the enum list */
 	PRI_MAX_TIMERS
 };
